"""Animations of pulsating sphere."""
import sfs
import numpy as np
from matplotlib import pyplot as plt
from matplotlib import animation


def particle_displacement(omega, center, radius, amplitude, grid, frames,
                          figsize=(8, 8), interval=80, blit=True, **kwargs):
    """Generate sound particle animation."""
    velocity = sfs.fd.source.pulsating_sphere_velocity(
               omega, center, radius, amplitude, grid)
    displacement = sfs.fd.displacement(velocity, omega)
    phasor = np.exp(1j * 2 * np.pi / frames)

    fig, ax = plt.subplots(figsize=figsize)
    ax.axis([grid[0].min(), grid[0].max(), grid[1].min(), grid[1].max()])
    scat = sfs.plot2d.particles(grid + displacement, **kwargs)

    def update_frame_displacement(i):
        position = (grid + displacement * phasor**i).apply(np.real)
        position = np.column_stack([position[0].flatten(),
                                    position[1].flatten()])
        scat.set_offsets(position)
        return [scat]

    return animation.FuncAnimation(
            fig, update_frame_displacement, frames,
            interval=interval, blit=blit)


def particle_velocity(omega, center, radius, amplitude, grid, frames,
                      figsize=(8, 8), interval=80, blit=True, **kwargs):
    """Generate particle velocity animation."""
    velocity = sfs.fd.source.pulsating_sphere_velocity(
               omega, center, radius, amplitude, grid)
    phasor = np.exp(1j * 2 * np.pi / frames)

    fig, ax = plt.subplots(figsize=figsize)
    ax.axis([grid[0].min(), grid[0].max(), grid[1].min(), grid[1].max()])
    quiv = sfs.plot2d.vectors(
            velocity, grid, clim=[-omega * amplitude, omega * amplitude],
            **kwargs)

    def update_frame_velocity(i):
        quiv.set_UVC(*(velocity[:2] * phasor**i).apply(np.real))
        return [quiv]

    return animation.FuncAnimation(
            fig, update_frame_velocity, frames, interval=interval, blit=True)


def sound_pressure(omega, center, radius, amplitude, grid, frames,
                   pulsate=False, figsize=(8, 8), interval=80, blit=True,
                   **kwargs):
    """Generate sound pressure animation."""
    pressure = sfs.fd.source.pulsating_sphere(
            omega, center, radius, amplitude, grid, inside=pulsate)
    phasor = np.exp(1j * 2 * np.pi / frames)

    fig, ax = plt.subplots(figsize=figsize)
    im = sfs.plot2d.amplitude(np.real(pressure), grid, **kwargs)
    ax.axis([grid[0].min(), grid[0].max(), grid[1].min(), grid[1].max()])

    def update_frame_pressure(i):
        distance = np.linalg.norm(grid)
        p = pressure * phasor**i
        if pulsate:
            p[distance <= radius + amplitude * np.real(phasor**i)] = np.nan
        im.set_array(np.real(p))
        return [im]

    return animation.FuncAnimation(
            fig, update_frame_pressure, frames, interval=interval, blit=True)


if __name__ == '__main__':

    # Pulsating sphere
    center = [0, 0, 0]
    radius = 0.25
    f = 750  # frequency
    omega = 2 * np.pi * f  # angular frequency

    # Axis limits
    xmin, xmax = -1, 1
    ymin, ymax = -1, 1

    # Animations
    frames = 20  # frames per period

    # Particle displacement
    amplitude = 5e-2  # amplitude of the surface displacement
    grid = sfs.util.xyz_grid([xmin, xmax], [ymin, ymax], 0, spacing=0.025)
    ani = particle_displacement(
            omega, center, radius, amplitude, grid, frames, c='Gray')
    ani.save('pulsating_sphere_displacement.gif', dpi=80, writer='imagemagick')

    # Particle velocity
    amplitude = 1e-3  # amplitude of the surface displacement
    grid = sfs.util.xyz_grid([xmin, xmax], [ymin, ymax], 0, spacing=0.04)
    ani = particle_velocity(
            omega, center, radius, amplitude, grid, frames)
    ani.save('pulsating_sphere_velocity.gif', dpi=80, writer='imagemagick')

    # Sound pressure
    amplitude = 1e-6  # amplitude of the surface displacement
    impedance_pw = sfs.default.rho0 * sfs.default.c
    max_pressure = omega * impedance_pw * amplitude
    grid = sfs.util.xyz_grid([xmin, xmax], [ymin, ymax], 0, spacing=0.005)
    ani = sound_pressure(
            omega, center, radius, amplitude, grid, frames, pulsate=True,
            colorbar=True, vmin=-max_pressure, vmax=max_pressure)
    ani.save('pulsating_sphere_pressure.gif', dpi=80, writer='imagemagick')



r"""Compute the sound field generated by a sound source.

.. include:: math-definitions.rst

.. plot::
    :context: reset

    import sfs
    import numpy as np
    import matplotlib.pyplot as plt
    plt.rcParams['figure.figsize'] = 8, 4.5  # inch

    x0 = 1.5, 1, 0
    f = 500  # Hz
    omega = 2 * np.pi * f

    normalization_point = 4 * np.pi
    normalization_line = \
        np.sqrt(8 * np.pi * omega / sfs.default.c) * np.exp(1j * np.pi / 4)

    grid = sfs.util.xyz_grid([-2, 3], [-1, 2], 0, spacing=0.02)

    # Grid for vector fields:
    vgrid = sfs.util.xyz_grid([-2, 3], [-1, 2], 0, spacing=0.1)

"""

import itertools
import numpy as np
from scipy import special
from .. import util
from .. import default


def point(omega, x0, grid, *, c=None):
    r"""Sound pressure of a point source.

    Parameters
    ----------
    omega : float
        Frequency of source.
    x0 : (3,) array_like
        Position of source.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    c : float, optional
        Speed of sound.

    Returns
    -------
    numpy.ndarray
        Sound pressure at positions given by *grid*.

    Notes
    -----
    .. math::

        G(\x-\x_0,\w) = \frac{1}{4\pi} \frac{\e{-\i\wc|\x-\x_0|}}{|\x-\x_0|}

    Examples
    --------
    .. plot::
        :context: close-figs

        p = sfs.fd.source.point(omega, x0, grid)
        sfs.plot2d.amplitude(p, grid)
        plt.title("Point Source at {} m".format(x0))

    Normalization ...

    .. plot::
        :context: close-figs

        sfs.plot2d.amplitude(p * normalization_point, grid,
                             colorbar_kwargs=dict(label="p / Pa"))
        plt.title("Point Source at {} m (normalized)".format(x0))

    """
    k = util.wavenumber(omega, c)
    x0 = util.asarray_1d(x0)
    grid = util.as_xyz_components(grid)

    r = np.linalg.norm(grid - x0)
    # If r is 0, the sound pressure is complex infinity
    numerator = np.exp(-1j * k * r) / (4*np.pi)
    with np.errstate(invalid='ignore', divide='ignore'):
        return numerator / r


def point_velocity(omega, x0, grid, *, c=None, rho0=None):
    """Particle velocity of a point source.

    Parameters
    ----------
    omega : float
        Frequency of source.
    x0 : (3,) array_like
        Position of source.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    c : float, optional
        Speed of sound.
    rho0 : float, optional
        Static density of air.

    Returns
    -------
    `XyzComponents`
        Particle velocity at positions given by *grid*.

    Examples
    --------
    The particle velocity can be plotted on top of the sound pressure:

    .. plot::
        :context: close-figs

        v = sfs.fd.source.point_velocity(omega, x0, vgrid)
        sfs.plot2d.amplitude(p * normalization_point, grid)
        sfs.plot2d.vectors(v * normalization_point, vgrid)
        plt.title("Sound Pressure and Particle Velocity")

    """
    if c is None:
        c = default.c
    if rho0 is None:
        rho0 = default.rho0
    k = util.wavenumber(omega, c)
    x0 = util.asarray_1d(x0)
    grid = util.as_xyz_components(grid)
    offset = grid - x0
    r = np.linalg.norm(offset)
    v = point(omega, x0, grid, c=c)
    v *= (1+1j*k*r) / (rho0 * c * 1j*k*r)
    return util.XyzComponents([v * o / r for o in offset])


def point_averaged_intensity(omega, x0, grid, *, c=None, rho0=None):
    """Velocity of a point source.

    Parameters
    ----------
    omega : float
        Frequency of source.
    x0 : (3,) array_like
        Position of source.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    c : float, optional
        Speed of sound.
    rho0 : float, optional
        Static density of air.

    Returns
    -------
    `XyzComponents`
        Averaged intensity at positions given by *grid*.

    """
    if c is None:
        c = default.c
    if rho0 is None:
        rho0 = default.rho0
    x0 = util.asarray_1d(x0)
    grid = util.as_xyz_components(grid)
    offset = grid - x0
    r = np.linalg.norm(offset)
    i = 1 / (2 * rho0 * c)
    return util.XyzComponents([i * o / r**2 for o in offset])


def point_dipole(omega, x0, n0, grid, *, c=None):
    r"""Point source with dipole characteristics.

    Parameters
    ----------
    omega : float
        Frequency of source.
    x0 : (3,) array_like
        Position of source.
    n0 : (3,) array_like
        Normal vector (direction) of dipole.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    c : float, optional
        Speed of sound.

    Returns
    -------
    numpy.ndarray
        Sound pressure at positions given by *grid*.

    Notes
    -----
    .. math::

        G(\x-\x_0,\w) = \frac{1}{4\pi} \left(\i\wc + \frac{1}{|\x-\x_0|}\right)
            \frac{\scalarprod{\x-\x_0}{\n_\text{s}}}{|\x-\x_0|^2}
            \e{-\i\wc|\x-\x_0}

    Examples
    --------
    .. plot::
        :context: close-figs

        n0 = 0, 1, 0
        p = sfs.fd.source.point_dipole(omega, x0, n0, grid)
        sfs.plot2d.amplitude(p, grid)
        plt.title("Dipole Point Source at {} m".format(x0))

    """
    k = util.wavenumber(omega, c)
    x0 = util.asarray_1d(x0)
    n0 = util.asarray_1d(n0)
    grid = util.as_xyz_components(grid)

    offset = grid - x0
    r = np.linalg.norm(offset)
    return 1 / (4*np.pi) * (1j * k + 1 / r) * np.inner(offset, n0) / \
        np.power(r, 2) * np.exp(-1j * k * r)


def point_modal(omega, x0, grid, L, *, N=None, deltan=0, c=None):
    """Point source in a rectangular room using a modal room model.

    Parameters
    ----------
    omega : float
        Frequency of source.
    x0 : (3,) array_like
        Position of source.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    L : (3,) array_like
        Dimensionons of the rectangular room.
    N : (3,) array_like or int, optional
        For all three spatial dimensions per dimension maximum order or
        list of orders. A scalar applies to all three dimensions. If no
        order is provided it is approximately determined.
    deltan : float, optional
        Absorption coefficient of the walls.
    c : float, optional
        Speed of sound.

    Returns
    -------
    numpy.ndarray
        Sound pressure at positions given by *grid*.

    """
    k = util.wavenumber(omega, c)
    x0 = util.asarray_1d(x0)
    x, y, z = util.as_xyz_components(grid)

    if np.isscalar(N):
        N = N * np.ones(3, dtype=int)

    if N is None:
            N = [None, None, None]

    orders = [0, 0, 0]
    for i in range(3):
        if N[i] is None:
            # compute max order
            orders[i] = range(int(np.ceil(L[i]/np.pi * k) + 1))
        elif np.isscalar(N[i]):
            # use given max order
            orders[i] = range(N[i] + 1)
        else:
            # use given orders
            orders[i] = N[i]

    kmp0 = [((kx + 1j * deltan)**2, np.cos(kx * x) * np.cos(kx * x0[0]))
            for kx in [m * np.pi / L[0] for m in orders[0]]]
    kmp1 = [((ky + 1j * deltan)**2, np.cos(ky * y) * np.cos(ky * x0[1]))
            for ky in [n * np.pi / L[1] for n in orders[1]]]
    kmp2 = [((kz + 1j * deltan)**2, np.cos(kz * z) * np.cos(kz * x0[2]))
            for kz in [l * np.pi / L[2] for l in orders[2]]]
    ksquared = k**2
    p = 0
    for (km0, p0), (km1, p1), (km2, p2) in itertools.product(kmp0, kmp1, kmp2):
        km = km0 + km1 + km2
        p = p + 8 / (ksquared - km) * p0 * p1 * p2
    return p


def point_modal_velocity(omega, x0, grid, L, *, N=None, deltan=0, c=None):
    """Velocity of point source in a rectangular room using a modal room model.

    Parameters
    ----------
    omega : float
        Frequency of source.
    x0 : (3,) array_like
        Position of source.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    L : (3,) array_like
        Dimensionons of the rectangular room.
    N : (3,) array_like or int, optional
        Combination of modal orders in the three-spatial dimensions to
        calculate the sound field for or maximum order for all
        dimensions.  If not given, the maximum modal order is
        approximately determined and the sound field is computed up to
        this maximum order.
    deltan : float, optional
        Absorption coefficient of the walls.
    c : float, optional
        Speed of sound.

    Returns
    -------
    `XyzComponents`
        Particle velocity at positions given by *grid*.

    """
    k = util.wavenumber(omega, c)
    x0 = util.asarray_1d(x0)
    x, y, z = util.as_xyz_components(grid)

    if N is None:
        # determine maximum modal order per dimension
        Nx = int(np.ceil(L[0]/np.pi * k))
        Ny = int(np.ceil(L[1]/np.pi * k))
        Nz = int(np.ceil(L[2]/np.pi * k))
        mm = range(Nx)
        nn = range(Ny)
        ll = range(Nz)
    elif np.isscalar(N):
        # compute up to a given order
        mm = range(N)
        nn = range(N)
        ll = range(N)
    else:
        # compute field for one order combination only
        mm = [N[0]]
        nn = [N[1]]
        ll = [N[2]]

    kmp0 = [((kx + 1j * deltan)**2, np.sin(kx * x) * np.cos(kx * x0[0]))
            for kx in [m * np.pi / L[0] for m in mm]]
    kmp1 = [((ky + 1j * deltan)**2, np.sin(ky * y) * np.cos(ky * x0[1]))
            for ky in [n * np.pi / L[1] for n in nn]]
    kmp2 = [((kz + 1j * deltan)**2, np.sin(kz * z) * np.cos(kz * x0[2]))
            for kz in [l * np.pi / L[2] for l in ll]]
    ksquared = k**2
    vx = 0+0j
    vy = 0+0j
    vz = 0+0j
    for (km0, p0), (km1, p1), (km2, p2) in itertools.product(kmp0, kmp1, kmp2):
        km = km0 + km1 + km2
        vx = vx - 8*1j / (ksquared - km) * p0
        vy = vy - 8*1j / (ksquared - km) * p1
        vz = vz - 8*1j / (ksquared - km) * p2
    return util.XyzComponents([vx, vy, vz])


def point_image_sources(omega, x0, grid, L, *, max_order, coeffs=None, c=None):
    """Point source in a rectangular room using the mirror image source model.

    Parameters
    ----------
    omega : float
        Frequency of source.
    x0 : (3,) array_like
        Position of source.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    L : (3,) array_like
        Dimensions of the rectangular room.
    max_order : int
        Maximum number of reflections for each image source.
    coeffs : (6,) array_like, optional
        Reflection coeffecients of the walls.
        If not given, the reflection coefficients are set to one.
    c : float, optional
        Speed of sound.

    Returns
    -------
    numpy.ndarray
        Sound pressure at positions given by *grid*.

    """
    if coeffs is None:
        coeffs = np.ones(6)

    xs, order = util.image_sources_for_box(x0, L, max_order)
    source_strengths = np.prod(coeffs**order, axis=1)

    p = 0
    for position, strength in zip(xs, source_strengths):
        if strength != 0:
            # point can be complex infinity
            with np.errstate(invalid='ignore'):
                p += strength * point(omega, position, grid, c=c)

    return p


def line(omega, x0, grid, *, c=None):
    r"""Line source parallel to the z-axis.

    Note: third component of x0 is ignored.

    Notes
    -----
    .. math::

        G(\x-\x_0,\w) = -\frac{\i}{4} \Hankel{2}{0}{\wc|\x-\x_0|}

    Examples
    --------
    .. plot::
        :context: close-figs

        p = sfs.fd.source.line(omega, x0, grid)
        sfs.plot2d.amplitude(p, grid)
        plt.title("Line Source at {} m".format(x0[:2]))

    Normalization ...

    .. plot::
        :context: close-figs

        sfs.plot2d.amplitude(p * normalization_line, grid,
                             colorbar_kwargs=dict(label="p / Pa"))
        plt.title("Line Source at {} m (normalized)".format(x0[:2]))

    """
    k = util.wavenumber(omega, c)
    x0 = util.asarray_1d(x0)[:2]  # ignore z-component
    grid = util.as_xyz_components(grid)

    r = np.linalg.norm(grid[:2] - x0)
    p = -1j/4 * _hankel2_0(k * r)
    return _duplicate_zdirection(p, grid)


def line_velocity(omega, x0, grid, *, c=None, rho0=None):
    """Velocity of line source parallel to the z-axis.

    Returns
    -------
    `XyzComponents`
        Particle velocity at positions given by *grid*.

    Examples
    --------
    The particle velocity can be plotted on top of the sound pressure:

    .. plot::
        :context: close-figs

        v = sfs.fd.source.line_velocity(omega, x0, vgrid)
        sfs.plot2d.amplitude(p * normalization_line, grid)
        sfs.plot2d.vectors(v * normalization_line, vgrid)
        plt.title("Sound Pressure and Particle Velocity")

    """
    if c is None:
        c = default.c
    if rho0 is None:
        rho0 = default.rho0
    k = util.wavenumber(omega, c)
    x0 = util.asarray_1d(x0)[:2]  # ignore z-component
    grid = util.as_xyz_components(grid)

    offset = grid[:2] - x0
    r = np.linalg.norm(offset)
    v = -1/(4 * c * rho0) * special.hankel2(1, k * r)
    v = [v * o / r for o in offset]

    assert v[0].shape == v[1].shape

    if len(grid) > 2:
        v.append(np.zeros_like(v[0]))

    return util.XyzComponents([_duplicate_zdirection(vi, grid) for vi in v])


def line_dipole(omega, x0, n0, grid, *, c=None):
    r"""Line source with dipole characteristics parallel to the z-axis.

    Note: third component of x0 is ignored.

    Notes
    -----
    .. math::

        G(\x-\x_0,\w) = \frac{\i k}{4} \Hankel{2}{1}{\wc|\x-\x_0|} \cos{\phi}

    """
    k = util.wavenumber(omega, c)
    x0 = util.asarray_1d(x0)[:2]  # ignore z-components
    n0 = util.asarray_1d(n0)[:2]
    grid = util.as_xyz_components(grid)
    dx = grid[:2] - x0

    r = np.linalg.norm(dx)
    p = 1j*k/4 * special.hankel2(1, k * r) * np.inner(dx, n0) / r
    return _duplicate_zdirection(p, grid)


def line_dirichlet_edge(omega, x0, grid, *, alpha=3/2*np.pi, Nc=None, c=None):
    """Line source scattered at an edge with Dirichlet boundary conditions.

    :cite:`Moser2012`, eq.(10.18/19)

    Parameters
    ----------
    omega : float
        Angular frequency.
    x0 : (3,) array_like
        Position of line source.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    alpha : float, optional
        Outer angle of edge.
    Nc : int, optional
        Number of elements for series expansion of driving function.
        Estimated if not given.
    c : float, optional
        Speed of sound

    Returns
    -------
    numpy.ndarray
        Complex pressure at grid positions.

    """
    k = util.wavenumber(omega, c)
    x0 = util.asarray_1d(x0)
    phi_s = np.arctan2(x0[1], x0[0])
    if phi_s < 0:
        phi_s = phi_s + 2*np.pi
    r_s = np.linalg.norm(x0)

    grid = util.XyzComponents(grid)

    r = np.linalg.norm(grid[:2])
    phi = np.arctan2(grid[1], grid[0])
    phi = np.where(phi < 0, phi+2*np.pi, phi)

    if Nc is None:
        Nc = np.ceil(2 * k * np.max(r) * alpha/np.pi)

    epsilon = np.ones(Nc)  # weights for series expansion
    epsilon[0] = 2

    p = np.zeros((grid[0].shape[1], grid[1].shape[0]), dtype=complex)
    idxr = (r <= r_s)
    idxa = (phi <= alpha)
    for m in np.arange(Nc):
        nu = m*np.pi/alpha
        f = 1/epsilon[m] * np.sin(nu*phi_s) * np.sin(nu*phi)
        p[idxr & idxa] = p[idxr & idxa] + f[idxr & idxa] * \
            special.jn(nu, k*r[idxr & idxa]) * special.hankel2(nu, k*r_s)
        p[~idxr & idxa] = p[~idxr & idxa] + f[~idxr & idxa] * \
            special.jn(nu, k*r_s) * special.hankel2(nu, k*r[~idxr & idxa])

    p = p * -1j*np.pi/alpha

    pl = line(omega, x0, None, grid, c=c)
    p[~idxa] = pl[~idxa]

    return p


def plane(omega, x0, n0, grid, *, c=None):
    r"""Plane wave.

    Parameters
    ----------
    omega : float
        Frequency of plane wave.
    x0 : (3,) array_like
        Position of plane wave.
    n0 : (3,) array_like
        Normal vector (direction) of plane wave.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    c : float, optional
        Speed of sound.

    Returns
    -------
    numpy.ndarray
        Sound pressure at positions given by *grid*.

    Notes
    -----
    .. math::

        G(\x,\w) = \e{-\i\wc\n\x}

    Examples
    --------
    .. plot::
        :context: close-figs

        direction = 45  # degree
        n0 = sfs.util.direction_vector(np.radians(direction))
        p = sfs.fd.source.plane(omega, x0, n0, grid)
        sfs.plot2d.amplitude(p, grid, colorbar_kwargs=dict(label="p / Pa"))
        plt.title("Plane wave with direction {} degree".format(direction))

    """
    k = util.wavenumber(omega, c)
    x0 = util.asarray_1d(x0)
    n0 = util.normalize_vector(n0)
    grid = util.as_xyz_components(grid)
    return np.exp(-1j * k * np.inner(grid - x0, n0))


def plane_velocity(omega, x0, n0, grid, *, c=None, rho0=None):
    r"""Velocity of a plane wave.

    Parameters
    ----------
    omega : float
        Frequency of plane wave.
    x0 : (3,) array_like
        Position of plane wave.
    n0 : (3,) array_like
        Normal vector (direction) of plane wave.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    c : float, optional
        Speed of sound.
    rho0 : float, optional
        Static density of air.

    Returns
    -------
    `XyzComponents`
        Particle velocity at positions given by *grid*.

    Notes
    -----
    .. math::

        V(\x,\w) = \frac{1}{\rho c} \e{-\i\wc\n\x} \n

    Examples
    --------
    The particle velocity can be plotted on top of the sound pressure:

    .. plot::
        :context: close-figs

        v = sfs.fd.source.plane_velocity(omega, x0, n0, vgrid)
        sfs.plot2d.amplitude(p, grid)
        sfs.plot2d.vectors(v, vgrid)
        plt.title("Sound Pressure and Particle Velocity")

    """
    if c is None:
        c = default.c
    if rho0 is None:
        rho0 = default.rho0
    v = plane(omega, x0, n0, grid, c=c) / (rho0 * c)
    return util.XyzComponents([v * n for n in n0])


def plane_averaged_intensity(omega, x0, n0, grid, *, c=None, rho0=None):
    r"""Averaged intensity of a plane wave.

    Parameters
    ----------
    omega : float
        Frequency of plane wave.
    x0 : (3,) array_like
        Position of plane wave.
    n0 : (3,) array_like
        Normal vector (direction) of plane wave.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    c : float, optional
        Speed of sound.
    rho0 : float, optional
        Static density of air.

    Returns
    -------
    `XyzComponents`
        Averaged intensity at positions given by *grid*.

    Notes
    -----
    .. math::

        I(\x,\w) = \frac{1}{2\rho c} \n

    """
    if c is None:
        c = default.c
    if rho0 is None:
        rho0 = default.rho0
    i = 1 / (2 * rho0 * c)
    return util.XyzComponents([i * n for n in n0])


def pulsating_sphere(omega, center, radius, amplitude, grid, *, inside=False,
                     c=None):
    """Sound pressure of a pulsating sphere.

    Parameters
    ---------
    omega : float
        Frequency of pulsating sphere
    center : (3,) array_like
        Center of sphere.
    radius : float
        Radius of sphere.
    amplitude : float
        Amplitude of displacement.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    inside : bool, optional
        As default, `numpy.nan` is returned for inside the sphere.
        If ``inside=True``, the sound field inside the sphere is extrapolated.
    c : float, optional
        Speed of sound.

    Returns
    -------
    numpy.ndarray
        Sound pressure at positions given by *grid*.
        If ``inside=False``, `numpy.nan` is returned for inside the sphere.

    Examples
    --------

    .. plot::
        :context: close-figs

        radius = 0.25
        amplitude = 1 / (radius * omega * sfs.default.rho0 * sfs.default.c)
        p = sfs.fd.source.pulsating_sphere(omega, x0, radius, amplitude, grid)
        sfs.plot2d.amplitude(p, grid)
        plt.title("Sound Pressure of a Pulsating Sphere")

    """
    if c is None:
        c = default.c
    k = util.wavenumber(omega, c)
    center = util.asarray_1d(center)
    grid = util.as_xyz_components(grid)

    distance = np.linalg.norm(grid - center)
    theta = np.arctan(1, k * distance)
    impedance = default.rho0 * c * np.cos(theta) * np.exp(1j * theta)
    radial_velocity = 1j * omega * amplitude * radius / distance \
        * np.exp(-1j * k * (distance - radius))
    if not inside:
        radial_velocity[distance <= radius] = np.nan
    return impedance * radial_velocity


def pulsating_sphere_velocity(omega, center, radius, amplitude, grid, *,
                              c=None):
    """Particle velocity of a pulsating sphere.

    Parameters
    ---------
    omega : float
        Frequency of pulsating sphere
    center : (3,) array_like
        Center of sphere.
    radius : float
        Radius of sphere.
    amplitude : float
        Amplitude of displacement.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    c : float, optional
        Speed of sound.

    Returns
    -------
    `XyzComponents`
        Particle velocity at positions given by *grid*.
        `numpy.nan` is returned for inside the sphere.

    Examples
    --------

    .. plot::
        :context: close-figs

        v = sfs.fd.source.pulsating_sphere_velocity(omega, x0, radius, amplitude, vgrid)
        sfs.plot2d.amplitude(p, grid)
        sfs.plot2d.vectors(v, vgrid)
        plt.title("Sound Pressure and Particle Velocity of a Pulsating Sphere")

    """
    if c is None:
        c = default.c
    k = util.wavenumber(omega, c)
    grid = util.as_xyz_components(grid)

    center = util.asarray_1d(center)
    offset = grid - center
    distance = np.linalg.norm(offset)
    radial_velocity = 1j * omega * amplitude * radius / distance \
        * np.exp(-1j * k * (distance - radius))
    radial_velocity[distance <= radius] = np.nan
    return util.XyzComponents([radial_velocity * o / distance for o in offset])


def _duplicate_zdirection(p, grid):
    """If necessary, duplicate field in z-direction."""
    gridshape = np.broadcast(*grid).shape
    if len(gridshape) > 2:
        return np.tile(p, [1, 1, gridshape[2]])
    else:
        return p


def _hankel2_0(x):
    """Wrapper for Hankel function of the second type using fast versions
       of the Bessel functions of first/second kind in scipy"""
    return special.j0(x)-1j*special.y0(x)



#!/usr/bin/env python3
from pathlib import Path
import subprocess
import sys

if __name__ != '__main__':
    raise ImportError(__name__ + ' is not meant be imported')

self = Path(__file__)
cwd = self.parent

for script in cwd.glob('*.py'):
    if self == script:
        # Don't call yourself!
        continue
    if script.name == 'ipython_kernel_config.py':
        # This is a configuration file, not an example script
        continue
    print('Running', script, '...')
    args = [sys.executable, str(script.relative_to(cwd))] + sys.argv[1:]
    result = subprocess.run(args, cwd=str(cwd))
    if result.returncode:
        print('Error running', script, file=sys.stderr)
        sys.exit(result.returncode)



"""Compute the sound field generated by a sound source.

The Green's function describes the spatial sound propagation over time.

.. include:: math-definitions.rst

.. plot::
    :context: reset

    import matplotlib.pyplot as plt
    import numpy as np
    from scipy.signal import unit_impulse
    import sfs

    xs = 1.5, 1, 0  # source position
    rs = np.linalg.norm(xs)  # distance from origin
    ts = rs / sfs.default.c  # time-of-arrival at origin

    # Impulsive excitation
    fs = 44100
    signal = unit_impulse(512), fs

    grid = sfs.util.xyz_grid([-2, 3], [-1, 2], 0, spacing=0.02)

"""

import numpy as np
from .. import util
from .. import default


def point(xs, signal, observation_time, grid, c=None):
    r"""Source model for a point source: 3D Green's function.

    Calculates the scalar sound pressure field for a given point in
    time, evoked by source excitation signal.

    Parameters
    ----------
    xs : (3,) array_like
        Position of source in cartesian coordinates.
    signal : (N,) array_like + float
        Excitation signal consisting of (mono) audio data and a sampling
        rate (in Hertz).  A `DelayedSignal` object can also be used.
    observation_time : float
        Observed point in time.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    c : float, optional
        Speed of sound.

    Returns
    -------
    numpy.ndarray
        Scalar sound pressure field, evaluated at positions given by
        *grid*.

    Notes
    -----
    .. math::

        g(x-x_s,t) = \frac{1}{4 \pi |x - x_s|} \dirac{t - \frac{|x -
        x_s|}{c}}

    Examples
    --------
    .. plot::
        :context: close-figs

        p = sfs.td.source.point(xs, signal, ts, grid)
        sfs.plot2d.level(p, grid)

    """
    xs = util.asarray_1d(xs)
    data, samplerate, signal_offset = util.as_delayed_signal(signal)
    data = util.asarray_1d(data)
    grid = util.as_xyz_components(grid)
    if c is None:
        c = default.c
    r = np.linalg.norm(grid - xs)
    # If r is +-0, the sound pressure is +-infinity
    with np.errstate(divide='ignore'):
        weights = 1 / (4 * np.pi * r)
    delays = r / c
    base_time = observation_time - signal_offset
    points_at_time = np.interp(base_time - delays,
                               np.arange(len(data)) / samplerate,
                               data, left=0, right=0)
    # weights can be +-infinity
    with np.errstate(invalid='ignore'):
        return weights * points_at_time


def point_image_sources(x0, signal, observation_time, grid, L, max_order,
                        coeffs=None, c=None):
    """Point source in a rectangular room using the mirror image source model.

    Parameters
    ----------
    x0 : (3,) array_like
        Position of source in cartesian coordinates.
    signal : (N,) array_like + float
        Excitation signal consisting of (mono) audio data and a sampling
        rate (in Hertz).  A `DelayedSignal` object can also be used.
    observation_time : float
        Observed point in time.
    grid : triple of array_like
        The grid that is used for the sound field calculations.
        See `sfs.util.xyz_grid()`.
    L : (3,) array_like
        Dimensions of the rectangular room.
    max_order : int
        Maximum number of reflections for each image source.
    coeffs : (6,) array_like, optional
        Reflection coeffecients of the walls.
        If not given, the reflection coefficients are set to one.
    c : float, optional
        Speed of sound.

    Returns
    -------
    numpy.ndarray
        Scalar sound pressure field, evaluated at positions given by
        *grid*.

    Examples
    --------
    .. plot::
        :context: close-figs

        room = 5, 3, 1.5  # room dimensions
        order = 2  # image source order
        coeffs = .8, .8, .6, .6, .7, .7  # wall reflection coefficients
        grid = sfs.util.xyz_grid([0, room[0]], [0, room[1]], 0, spacing=0.01)
        p = sfs.td.source.point_image_sources(
                xs, signal, 1.5 * ts, grid, room, order, coeffs)
        sfs.plot2d.level(p, grid)

    """
    if coeffs is None:
        coeffs = np.ones(6)

    positions, order = util.image_sources_for_box(x0, L, max_order)
    source_strengths = np.prod(coeffs**order, axis=1)

    p = 0
    for position, strength in zip(positions, source_strengths):
        if strength != 0:
            p += strength * point(position, signal, observation_time, grid, c)

    return p



"""3D plots of sound fields etc."""
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


def secondary_sources(x0, n0, a0=None, *, w=0.08, h=0.08):
    """Plot positions and normals of a 3D secondary source distribution."""
    fig = plt.figure(figsize=(15, 15))
    ax = fig.add_subplot(111, projection='3d')
    q = ax.quiver(x0[:, 0], x0[:, 1], x0[:, 2], n0[:, 0],
                  n0[:, 1], n0[:, 2], length=0.1)
    plt.xlabel('x (m)')
    plt.ylabel('y (m)')
    plt.title('Secondary Sources')
    return q



import numpy as np
from numpy.testing import assert_allclose
import pytest
import sfs


cart_sph_data = [
    ((1, 1, 1), (np.pi / 4, np.arccos(1 / np.sqrt(3)), np.sqrt(3))),
    ((-1, 1, 1), (3 / 4 * np.pi, np.arccos(1 / np.sqrt(3)), np.sqrt(3))),
    ((1, -1, 1), (-np.pi / 4, np.arccos(1 / np.sqrt(3)), np.sqrt(3))),
    ((-1, -1, 1), (-3 / 4 * np.pi, np.arccos(1 / np.sqrt(3)), np.sqrt(3))),
    ((1, 1, -1), (np.pi / 4, np.arccos(-1 / np.sqrt(3)), np.sqrt(3))),
    ((-1, 1, -1), (3 / 4 * np.pi, np.arccos(-1 / np.sqrt(3)), np.sqrt(3))),
    ((1, -1, -1), (-np.pi / 4, np.arccos(-1 / np.sqrt(3)), np.sqrt(3))),
    ((-1, -1, -1), (-3 / 4 * np.pi, np.arccos(-1 / np.sqrt(3)), np.sqrt(3))),
]


@pytest.mark.parametrize('coord, polar', cart_sph_data)
def test_cart2sph(coord, polar):
    x, y, z = coord
    a = sfs.util.cart2sph(x, y, z)
    assert_allclose(a, polar)


@pytest.mark.parametrize('coord, polar', cart_sph_data)
def test_sph2cart(coord, polar):
    alpha, beta, r = polar
    b = sfs.util.sph2cart(alpha, beta, r)
    assert_allclose(b, coord)


direction_vector_data = [
    ((np.pi / 4, np.pi / 4), (0.5, 0.5, np.sqrt(2) / 2)),
    ((3 * np.pi / 4, 3 * np.pi / 4), (-1 / 2, 1 / 2, -np.sqrt(2) / 2)),
    ((3 * np.pi / 4, -3 * np.pi / 4), (1 / 2, -1 / 2, -np.sqrt(2) / 2)),
    ((np.pi / 4, -np.pi / 4), (-1 / 2, -1 / 2, np.sqrt(2) / 2)),
    ((-np.pi / 4, np.pi / 4), (1 / 2, -1 / 2, np.sqrt(2) / 2)),
    ((-3 * np.pi / 4, 3 * np.pi / 4), (-1 / 2, -1 / 2, -np.sqrt(2) / 2)),
    ((-3 * np.pi / 4, -3 * np.pi / 4), (1 / 2, 1 / 2, -np.sqrt(2) / 2)),
    ((-np.pi / 4, -np.pi / 4), (-1 / 2, 1 / 2, np.sqrt(2) / 2)),
]


@pytest.mark.parametrize('input, vector', direction_vector_data)
def test_direction_vector(input, vector):
    alpha, beta = input
    c = sfs.util.direction_vector(alpha, beta)
    assert_allclose(c, vector)


db_data = [
    (0, -np.inf),
    (0.5, -3.01029995663981),
    (1, 0),
    (2, 3.01029995663981),
    (10, 10),
]


@pytest.mark.parametrize('linear, power_db', db_data)
def test_db_amplitude(linear, power_db):
    d = sfs.util.db(linear)
    assert_allclose(d, power_db * 2)


@pytest.mark.parametrize('linear, power_db', db_data)
def test_db_power(linear, power_db):
    d = sfs.util.db(linear, power=True)
    assert_allclose(d, power_db)



"""2D plots of sound fields etc."""
import matplotlib.pyplot as plt
from matplotlib import __version__ as matplotlib_version
from matplotlib.patches import PathPatch
from matplotlib.path import Path
from matplotlib.collections import PatchCollection
from mpl_toolkits import axes_grid1
import numpy as np
from . import util
from . import default


def _register_cmap_clip(name, original_cmap, alpha):
    """Create a color map with "over" and "under" values."""
    from matplotlib.colors import LinearSegmentedColormap
    cdict = plt.cm.datad[original_cmap]
    cmap = LinearSegmentedColormap(name, cdict)
    cmap.set_over([alpha * c + 1 - alpha for c in cmap(1.0)[:3]])
    cmap.set_under([alpha * c + 1 - alpha for c in cmap(0.0)[:3]])
    plt.cm.register_cmap(cmap=cmap)


# The 'coolwarm' colormap is based on the paper
# "Diverging Color Maps for Scientific Visualization" by Kenneth Moreland
# http://www.sandia.gov/~kmorel/documents/ColorMaps/
_register_cmap_clip('coolwarm_clip', 'coolwarm', 0.7)


def _register_cmap_transparent(name, color):
    """Create a color map from a given color to transparent."""
    from matplotlib.colors import colorConverter, LinearSegmentedColormap
    red, green, blue = colorConverter.to_rgb(color)
    cdict = {'red': ((0, red, red), (1, red, red)),
             'green': ((0, green, green), (1, green, green)),
             'blue': ((0, blue, blue), (1, blue, blue)),
             'alpha': ((0, 0, 0), (1, 1, 1))}
    cmap = LinearSegmentedColormap(name, cdict)
    plt.cm.register_cmap(cmap=cmap)


_register_cmap_transparent('blacktransparent', 'black')


def virtualsource(xs, ns=None, type='point', *, ax=None):
    """Draw position/orientation of virtual source."""
    xs = np.asarray(xs)
    ns = np.asarray(ns)
    if ax is None:
        ax = plt.gca()

    if type == 'point':
        vps = plt.Circle(xs, .05, edgecolor='k', facecolor='k')
        ax.add_artist(vps)
        for n in range(1, 3):
            vps = plt.Circle(xs, .05+n*0.05, edgecolor='k', fill=False)
            ax.add_artist(vps)
    elif type == 'plane':
        ns = 0.2 * ns

        ax.arrow(xs[0], xs[1], ns[0], ns[1], head_width=0.05,
                 head_length=0.1, fc='k', ec='k')


def reference(xref, *, size=0.1, ax=None):
    """Draw reference/normalization point."""
    xref = np.asarray(xref)
    if ax is None:
        ax = plt.gca()

    ax.plot((xref[0]-size, xref[0]+size), (xref[1]-size, xref[1]+size), 'k-')
    ax.plot((xref[0]-size, xref[0]+size), (xref[1]+size, xref[1]-size), 'k-')


def secondary_sources(x0, n0, *, grid=None):
    """Simple plot of secondary source locations."""
    x0 = np.asarray(x0)
    n0 = np.asarray(n0)
    ax = plt.gca()

    # plot only secondary sources inside simulated area
    if grid is not None:
        x0, n0 = _visible_secondarysources(x0, n0, grid)

    # plot symbols
    for x00 in x0:
        ss = plt.Circle(x00[0:2], .05, edgecolor='k', facecolor='k')
        ax.add_artist(ss)


def loudspeakers(x0, n0, a0=0.5, *, size=0.08, show_numbers=False, grid=None,
                 ax=None):
    """Draw loudspeaker symbols at given locations and angles.

    Parameters
    ----------
    x0 : (N, 3) array_like
        Loudspeaker positions.
    n0 : (N, 3) or (3,) array_like
        Normal vector(s) of loudspeakers.
    a0 : float or (N,) array_like, optional
        Weighting factor(s) of loudspeakers.
    size : float, optional
        Size of loudspeakers in metres.
    show_numbers : bool, optional
        If ``True``, loudspeaker numbers are shown.
    grid : triple of array_like, optional
        If specified, only loudspeakers within the *grid* are shown.
    ax : Axes object, optional
        The loudspeakers are plotted into this `matplotlib.axes.Axes`
        object or -- if not specified -- into the current axes.

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    a0 = util.asarray_1d(a0).reshape(-1, 1)

    # plot only secondary sources inside simulated area
    if grid is not None:
        x0, n0 = _visible_secondarysources(x0, n0, grid)

    # normalized coordinates of loudspeaker symbol (see IEC 60617-9)
    codes, coordinates = zip(*(
        (Path.MOVETO, [-0.62, 0.21]),
        (Path.LINETO, [-0.31, 0.21]),
        (Path.LINETO, [0, 0.5]),
        (Path.LINETO, [0, -0.5]),
        (Path.LINETO, [-0.31, -0.21]),
        (Path.LINETO, [-0.62, -0.21]),
        (Path.CLOSEPOLY, [0, 0]),
        (Path.MOVETO, [-0.31, 0.21]),
        (Path.LINETO, [-0.31, -0.21]),
    ))
    coordinates = np.column_stack([coordinates, np.zeros(len(coordinates))])
    coordinates *= size

    patches = []
    for x00, n00 in util.broadcast_zip(x0, n0):
        # rotate and translate coordinates
        R = util.rotation_matrix([1, 0, 0], n00)
        transformed_coordinates = np.inner(coordinates, R) + x00

        patches.append(PathPatch(Path(transformed_coordinates[:, :2], codes)))

    # add collection of patches to current axis
    p = PatchCollection(patches, edgecolor='0', facecolor=np.tile(1 - a0, 3))
    if ax is None:
        ax = plt.gca()
    ax.add_collection(p)

    if show_numbers:
        for idx, (x00, n00) in enumerate(util.broadcast_zip(x0, n0)):
            x, y = x00[:2] - 1.2 * size * n00[:2]
            ax.text(x, y, idx + 1, horizontalalignment='center',
                    verticalalignment='center', clip_on=True)


def _visible_secondarysources(x0, n0, grid):
    """Determine secondary sources which lie within *grid*."""
    x, y = util.as_xyz_components(grid[:2])
    idx = np.where((x0[:, 0] > x.min()) & (x0[:, 0] < x.max()) &
                   (x0[:, 1] > y.min()) & (x0[:, 1] < x.max()))
    idx = np.squeeze(idx)

    return x0[idx, :], n0[idx, :]


def amplitude(p, grid, *, xnorm=None, cmap='coolwarm_clip',
              vmin=-2.0, vmax=2.0, xlabel=None, ylabel=None,
              colorbar=True, colorbar_kwargs={}, ax=None, **kwargs):
    """Two-dimensional plot of sound field (real part).

    Parameters
    ----------
    p : array_like
        Sound pressure values (or any other scalar quantity if you
        like).  If the values are complex, the imaginary part is
        ignored.
        Typically, *p* is two-dimensional with a shape of *(Ny, Nx)*,
        *(Nz, Nx)* or *(Nz, Ny)*.  This is the case if
        `sfs.util.xyz_grid()` was used with a single number for *z*,
        *y* or *x*, respectively.
        However, *p* can also be three-dimensional with a shape of *(Ny,
        Nx, 1)*, *(1, Nx, Nz)* or *(Ny, 1, Nz)*.  This is the case if
        :func:`numpy.meshgrid` was used with a scalar for *z*, *y* or
        *x*, respectively (and of course with the default
        ``indexing='xy'``).

        .. note:: If you want to plot a single slice of a pre-computed
                  "full" 3D sound field, make sure that the slice still
                  has three dimensions (including one singleton
                  dimension).  This way, you can use the original *grid*
                  of the full volume without changes.
                  This works because the grid component corresponding to
                  the singleton dimension is simply ignored.

    grid : triple or pair of numpy.ndarray
        The grid that was used to calculate *p*, see
        `sfs.util.xyz_grid()`.  If *p* is two-dimensional, but
        *grid* has 3 components, one of them must be scalar.
    xnorm : array_like, optional
        Coordinates of a point to which the sound field should be
        normalized before plotting.  If not specified, no normalization
        is used.  See `sfs.util.normalize()`.

    Returns
    -------
    AxesImage
        See :func:`matplotlib.pyplot.imshow`.

    Other Parameters
    ----------------
    xlabel, ylabel : str
        Overwrite default x/y labels.  Use ``xlabel=''`` and
        ``ylabel=''`` to remove x/y labels.  The labels can be changed
        afterwards with :func:`matplotlib.pyplot.xlabel` and
        :func:`matplotlib.pyplot.ylabel`.
    colorbar : bool, optional
        If ``False``, no colorbar is created.
    colorbar_kwargs : dict, optional
        Further colorbar arguments, see `add_colorbar()`.
    ax : Axes, optional
        If given, the plot is created on *ax* instead of the current
        axis (see :func:`matplotlib.pyplot.gca`).
    cmap, vmin, vmax, **kwargs
        All further parameters are forwarded to
        :func:`matplotlib.pyplot.imshow`.

    See Also
    --------
    sfs.plot2d.level

    """
    p = np.asarray(p)
    grid = util.as_xyz_components(grid)

    # normalize sound field wrt xnorm
    if xnorm is not None:
        p = util.normalize(p, grid, xnorm)

    if p.ndim == 3:
        if p.shape[2] == 1:
            p = p[:, :, 0]  # first axis: y; second axis: x
            plotting_plane = 'xy'
        elif p.shape[1] == 1:
            p = p[:, 0, :].T  # first axis: z; second axis: y
            plotting_plane = 'yz'
        elif p.shape[0] == 1:
            p = p[0, :, :].T  # first axis: z; second axis: x
            plotting_plane = 'xz'
        else:
            raise ValueError("If p is 3D, one dimension must have length 1")
    elif len(grid) == 3:
        if grid[2].ndim == 0:
            plotting_plane = 'xy'
        elif grid[1].ndim == 0:
            plotting_plane = 'xz'
        elif grid[0].ndim == 0:
            plotting_plane = 'yz'
        else:
            raise ValueError(
                "If p is 2D and grid is 3D, one grid component must be scalar")
    else:
        # 2-dimensional case
        plotting_plane = 'xy'

    if plotting_plane == 'xy':
        x, y = grid[[0, 1]]
    elif plotting_plane == 'xz':
        x, y = grid[[0, 2]]
    elif plotting_plane == 'yz':
        x, y = grid[[1, 2]]

    dx = 0.5 * x.ptp() / p.shape[0]
    dy = 0.5 * y.ptp() / p.shape[1]

    if ax is None:
        ax = plt.gca()

    # see https://github.com/matplotlib/matplotlib/issues/10567
    if matplotlib_version.startswith('2.1.'):
        p = np.clip(p, -1e15, 1e15)  # clip to float64 range

    im = ax.imshow(np.real(p), cmap=cmap, origin='lower',
                   extent=[x.min()-dx, x.max()+dx, y.min()-dy, y.max()+dy],
                   vmax=vmax, vmin=vmin, **kwargs)
    if xlabel is None:
        xlabel = plotting_plane[0] + ' / m'
    if ylabel is None:
        ylabel = plotting_plane[1] + ' / m'
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if colorbar:
        add_colorbar(im, **colorbar_kwargs)
    return im


def level(p, grid, *, xnorm=None, power=False, cmap=None, vmax=3, vmin=-50,
          **kwargs):
    """Two-dimensional plot of level (dB) of sound field.

    Takes the same parameters as `sfs.plot2d.amplitude()`.

    Other Parameters
    ----------------
    power : bool, optional
        See `sfs.util.db()`.

    """
    # normalize before converting to dB!
    if xnorm is not None:
        p = util.normalize(p, grid, xnorm)
    L = util.db(p, power=power)
    return amplitude(L, grid=grid, xnorm=None, cmap=cmap,
                     vmax=vmax, vmin=vmin, **kwargs)


def particles(x, *, trim=None, ax=None, xlabel='x (m)', ylabel='y (m)',
              edgecolor='', marker='.', s=15, **kwargs):
    """Plot particle positions as scatter plot"""
    XX, YY = [np.real(c) for c in x[:2]]

    if trim is not None:
        xmin, xmax, ymin, ymax = trim

        idx = np.where((XX > xmin) & (XX < xmax) & (YY > ymin) & (YY < ymax))
        XX = XX[idx]
        YY = YY[idx]

    if ax is None:
        ax = plt.gca()

    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)
    return ax.scatter(XX, YY, edgecolor=edgecolor, marker=marker, s=s,
                      **kwargs)


def vectors(v, grid, *, cmap='blacktransparent', headlength=3,
            headaxislength=2.5, ax=None, clim=None, **kwargs):
    """Plot a vector field in the xy plane.

    Parameters
    ----------
    v : triple or pair of array_like
        x, y and optionally z components of vector field.  The z
        components are ignored.
        If the values are complex, the imaginary parts are ignored.
    grid : triple or pair of array_like
        The grid that was used to calculate *v*, see
        `sfs.util.xyz_grid()`.  Any z components are ignored.

    Returns
    -------
    Quiver
        See :func:`matplotlib.pyplot.quiver`.

    Other Parameters
    ----------------
    ax : Axes, optional
        If given, the plot is created on *ax* instead of the current
        axis (see :func:`matplotlib.pyplot.gca`).
    clim : pair of float, optional
        Limits for the scaling of arrow colors.
        See :func:`matplotlib.pyplot.quiver`.
    cmap, headlength, headaxislength, **kwargs
        All further parameters are forwarded to
        :func:`matplotlib.pyplot.quiver`.

    """
    v = util.as_xyz_components(v[:2]).apply(np.real)
    X, Y = util.as_xyz_components(grid[:2])
    speed = np.linalg.norm(v)
    with np.errstate(invalid='ignore'):
        U, V = v.apply(np.true_divide, speed)
    if ax is None:
        ax = plt.gca()
    if clim is None:
        v_ref = 1 / (default.rho0 * default.c)  # reference particle velocity
        clim = 0, 2 * v_ref
    return ax.quiver(X, Y, U, V, speed, cmap=cmap, pivot='mid', units='xy',
                     angles='xy', headlength=headlength,
                     headaxislength=headaxislength, clim=clim, **kwargs)


def add_colorbar(im, *, aspect=20, pad=0.5, **kwargs):
    r"""Add a vertical color bar to a plot.

    Parameters
    ----------
    im : ScalarMappable
        The output of `sfs.plot2d.amplitude()`, `sfs.plot2d.level()` or any
        other `matplotlib.cm.ScalarMappable`.
    aspect : float, optional
        Aspect ratio of the colorbar.  Strictly speaking, since the
        colorbar is vertical, it's actually the inverse of the aspect
        ratio.
    pad : float, optional
        Space between image plot and colorbar, as a fraction of the
        width of the colorbar.

        .. note:: The *pad* argument of
                  :meth:`matplotlib.figure.Figure.colorbar` has a
                  slightly different meaning ("fraction of original
                  axes")!
    \**kwargs
        All further arguments are forwarded to
        :meth:`matplotlib.figure.Figure.colorbar`.

    See Also
    --------
    matplotlib.pyplot.colorbar

    """
    ax = im.axes
    divider = axes_grid1.make_axes_locatable(ax)
    width = axes_grid1.axes_size.AxesY(ax, aspect=1/aspect)
    pad = axes_grid1.axes_size.Fraction(pad, width)
    current_ax = plt.gca()
    cax = divider.append_axes("right", size=width, pad=pad)
    plt.sca(current_ax)
    return ax.figure.colorbar(im, cax=cax, orientation='vertical', **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SFS documentation build configuration file, created by
# sphinx-quickstart on Tue Nov  4 14:01:37 2014.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
from subprocess import check_output

import sphinx

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('..'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.3'  # for sphinx.ext.napoleon

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',  # support for NumPy-style docstrings
    'sphinx.ext.intersphinx',
    'sphinx.ext.doctest',
    'sphinxcontrib.bibtex',
    'sphinx.ext.extlinks',
    'matplotlib.sphinxext.plot_directive',
    'nbsphinx',
]

nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'svg', 'pdf'}",
    "--InlineBackend.rc={'figure.dpi': 96}",
]

# Tell autodoc that the documentation is being generated
sphinx.SFS_DOCS_ARE_BEING_BUILT = True

autoclass_content = 'init'
autodoc_member_order = 'bysource'
autodoc_default_flags = ['members', 'undoc-members']

autosummary_generate = ['api']

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = False
napoleon_use_rtype = False

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'matplotlib': ('https://matplotlib.org/', None),
}

extlinks = {'sfs': ('https://sfs.readthedocs.io/en/3.2/%s',
                    'https://sfs.rtfd.io/')}

plot_include_source = True
plot_html_show_source_link = False
plot_html_show_formats = False
plot_pre_code = ''
plot_rcparams = {
    'savefig.bbox': 'tight',
}
plot_formats = ['svg', 'pdf']

mathjax_config = {
    'TeX': {
        'extensions': ['newcommand.js', 'begingroup.js'],  # Support for \gdef
    },
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_template']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
authors = 'SFS Toolbox Developers'
project = 'SFS Toolbox'
copyright = '2019, ' + authors

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
#version = '0.0.0'
# The full version, including alpha/beta/rc tags.
try:
    release = check_output(['git', 'describe', '--tags', '--always'])
    release = release.decode().strip()
except Exception:
    release = '<unknown>'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'
try:
    today = check_output(['git', 'show', '-s', '--format=%ad', '--date=short'])
    today = today.decode().strip()
except Exception:
    today = '<unknown date>'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', '**/.ipynb_checkpoints']

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'any'

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

jinja_define = """
{% set docname = env.doc2path(env.docname, base='doc') %}
{% set latex_href = ''.join([
    '\href{https://github.com/sfstoolbox/sfs-python/blob/',
    env.config.release,
    '/',
    docname | escape_latex,
    '}{\sphinxcode{\sphinxupquote{',
    docname | escape_latex,
    '}}}',
]) %}
"""

nbsphinx_prolog = jinja_define + r"""
.. only:: html

    .. role:: raw-html(raw)
        :format: html

    .. nbinfo::

        This page was generated from `{{ docname }}`__.
        Interactive online version:
        :raw-html:`<a href="https://mybinder.org/v2/gh/sfstoolbox/sfs-python/{{ env.config.release }}?filepath={{ docname }}"><img alt="Binder badge" src="https://mybinder.org/badge_logo.svg" style="vertical-align:text-bottom"></a>`

    __ https://github.com/sfstoolbox/sfs-python/blob/
        {{ env.config.release }}/{{ docname }}

.. raw:: latex

    \nbsphinxstartnotebook{\scriptsize\noindent\strut
    \textcolor{gray}{The following section was generated from {{ latex_href }}
    \dotfill}}
"""

nbsphinx_epilog = jinja_define + r"""
.. raw:: latex

    \nbsphinxstopnotebook{\scriptsize\noindent\strut
    \textcolor{gray}{\dotfill\ {{ latex_href }} ends here.}}
"""


# -- Options for HTML output ----------------------------------------------

def setup(app):
    """Include custom theme files to sphinx HTML header"""
    app.add_stylesheet('css/title.css')

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    'collapse_navigation': False,
}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = project + ", version " + release

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True
html_sourcelink_suffix = ''

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'SFS'

html_scaled_image_link = False

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    'papersize': 'a4paper',
    'printindex': '',
    'sphinxsetup': r"""
        VerbatimColor={HTML}{F5F5F5},
        VerbatimBorderColor={HTML}{E0E0E0},
        noteBorderColor={HTML}{E0E0E0},
        noteborder=1.5pt,
        warningBorderColor={HTML}{E0E0E0},
        warningborder=1.5pt,
        warningBgColor={HTML}{FBFBFB},
    """,
    'preamble': r"""
\usepackage[sc,osf]{mathpazo}
\linespread{1.05}  % see http://www.tug.dk/FontCatalogue/urwpalladio/
\renewcommand{\sfdefault}{pplj}  % Palatino instead of sans serif
\IfFileExists{zlmtt.sty}{
    \usepackage[light,scaled=1.05]{zlmtt}  % light typewriter font from lmodern
}{
    \renewcommand{\ttdefault}{lmtt}  % typewriter font from lmodern
}
""",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [('index', 'SFS.tex', project, authors, 'howto')]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
latex_show_urls = 'footnote'

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
latex_domain_indices = False


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
#man_pages = [('index', 'sfs', project, [authors], 1)]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
#texinfo_documents = [
#  ('index', 'SFS', project, project, 'SFS', 'Sound Field Synthesis Toolbox.',
#   'Miscellaneous'),
#]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False


# -- Options for epub output ----------------------------------------------

epub_author = authors



"""Create some examples of time-domain NFC-HOA."""

import numpy as np
import matplotlib.pyplot as plt
import sfs
from scipy.signal import unit_impulse

# Parameters
fs = 44100  # sampling frequency
grid = sfs.util.xyz_grid([-2, 2], [-2, 2], 0, spacing=0.005)
N = 60  # number of secondary sources
R = 1.5  # radius of circular array
array = sfs.array.circular(N, R)

# Excitation signal
signal = unit_impulse(512), fs, 0

# Plane wave
max_order = None
npw = [0, -1, 0]  # propagating direction
t = 0  # observation time
delay, weight, sos, phaseshift, selection, secondary_source = \
    sfs.td.nfchoa.plane_25d(array.x, R, npw, fs, max_order)
d = sfs.td.nfchoa.driving_signals_25d(
        delay, weight, sos, phaseshift, signal)
p = sfs.td.synthesize(d, selection, array, secondary_source,
                      observation_time=t, grid=grid)

plt.figure()
sfs.plot2d.level(p, grid)
sfs.plot2d.loudspeakers(array.x, array.n)
sfs.plot2d.virtualsource([0, 0], ns=npw, type='plane')
plt.savefig('impulse_pw_nfchoa_25d.png')

# Point source
max_order = 100
xs = [1.5, 1.5, 0]  # position
t = np.linalg.norm(xs) / sfs.default.c  # observation time
delay, weight, sos, phaseshift, selection, secondary_source = \
    sfs.td.nfchoa.point_25d(array.x, R, xs, fs, max_order)
d = sfs.td.nfchoa.driving_signals_25d(
        delay, weight, sos, phaseshift, signal)
p = sfs.td.synthesize(d, selection, array, secondary_source,
                      observation_time=t, grid=grid)

plt.figure()
sfs.plot2d.level(p, grid)
sfs.plot2d.loudspeakers(array.x, array.n)
sfs.plot2d.virtualsource(xs, type='point')
plt.savefig('impulse_ps_nfchoa_25d.png')



"""
Create some examples in the time domain.

Simulate and plot impulse behavior for Wave Field Synthesis.

"""

import numpy as np
import matplotlib.pyplot as plt
import sfs

# simulation parameters
grid = sfs.util.xyz_grid([-3, 3], [-3, 3], 0, spacing=0.01)
my_cmap = 'YlOrRd'
N = 56  # number of secondary sources
R = 1.5  # radius of spherical/circular array
array = sfs.array.circular(N, R)  # get secondary source positions
fs = 44100  # sampling rate

# unit impulse
signal = [1], fs

# POINT SOURCE
xs = 2, 2, 0  # position of virtual source
t = 0.008
# compute driving signals
d_delay, d_weight, selection, secondary_source = \
    sfs.td.wfs.point_25d(array.x, array.n, xs)
d = sfs.td.wfs.driving_signals(d_delay, d_weight, signal)

# test soundfield
twin = sfs.tapering.tukey(selection, alpha=0.3)

p = sfs.td.synthesize(d, twin, array,
                      secondary_source, observation_time=t, grid=grid)
p = p * 100  # scale absolute amplitude

plt.figure(figsize=(10, 10))
sfs.plot2d.level(p, grid, cmap=my_cmap)
sfs.plot2d.loudspeakers(array.x, array.n, twin)
plt.grid()
sfs.plot2d.virtualsource(xs)
plt.title('impulse_ps_wfs_25d')
plt.savefig('impulse_ps_wfs_25d.png')

# PLANE WAVE
pw_angle = 30  # traveling direction of plane wave
npw = sfs.util.direction_vector(np.radians(pw_angle))
t = -0.001

# compute driving signals
d_delay, d_weight, selection, secondary_source = \
    sfs.td.wfs.plane_25d(array.x, array.n, npw)
d = sfs.td.wfs.driving_signals(d_delay, d_weight, signal)

# test soundfield
twin = sfs.tapering.tukey(selection, alpha=0.3)
p = sfs.td.synthesize(d, twin, array,
                      secondary_source, observation_time=t, grid=grid)

plt.figure(figsize=(10, 10))
sfs.plot2d.level(p, grid, cmap=my_cmap)
sfs.plot2d.loudspeakers(array.x, array.n, twin)
plt.grid()
sfs.plot2d.virtualsource([0, 0], npw, type='plane')
plt.title('impulse_pw_wfs_25d')
plt.savefig('impulse_pw_wfs_25d.png')

# FOCUSED SOURCE
xs = np.r_[0.5, 0.5, 0]  # position of virtual source
xref = np.r_[0, 0, 0]
nfs = sfs.util.normalize_vector(xref - xs)  # main n of fsource
t = 0.003  # compute driving signals
d_delay, d_weight, selection, secondary_source = \
    sfs.td.wfs.focused_25d(array.x, array.n, xs, nfs)
d = sfs.td.wfs.driving_signals(d_delay, d_weight, signal)

# test soundfield
twin = sfs.tapering.tukey(selection, alpha=0.3)
p = sfs.td.synthesize(d, twin, array,
                      secondary_source, observation_time=t, grid=grid)
p = p * 100  # scale absolute amplitude

plt.figure(figsize=(10, 10))
sfs.plot2d.level(p, grid, cmap=my_cmap)
sfs.plot2d.loudspeakers(array.x, array.n, twin)
plt.grid()
sfs.plot2d.virtualsource(xs)
plt.title('impulse_fs_wfs_25d')
plt.savefig('impulse_fs_wfs_25d.png')



from setuptools import setup, find_packages

__version__ = "unknown"

# "import" __version__
for line in open("sfs/__init__.py"):
    if line.startswith("__version__"):
        exec(line)
        break

setup(
    name="sfs",
    version=__version__,
    packages=find_packages(),
    install_requires=[
        'numpy!=1.11.0',  # https://github.com/sfstoolbox/sfs-python/issues/11
        'scipy',
    ],
    author="SFS Toolbox Developers",
    author_email="sfstoolbox@gmail.com",
    description="Sound Field Synthesis Toolbox",
    long_description=open('README.rst').read(),
    license="MIT",
    keywords="audio SFS WFS Ambisonics".split(),
    url="http://github.com/sfstoolbox/",
    platforms='any',
    python_requires='>=3.5',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering",
    ],
    zip_safe=True,
)



"""Sound Field Synthesis Toolbox.

https://sfs-python.readthedocs.io/

.. rubric:: Submodules

.. autosummary::
    :toctree:

    fd
    td
    array
    tapering
    plot2d
    plot3d
    util

"""
__version__ = "0.5.0"


class default:
    """Get/set defaults for the *sfs* module.

    For example, when you want to change the default speed of sound::

        import sfs
        sfs.default.c = 330

    """

    c = 343
    """Speed of sound."""

    rho0 = 1.2250
    """Static density of air."""

    selection_tolerance = 1e-6
    """Tolerance used for secondary source selection."""

    def __setattr__(self, name, value):
        """Only allow setting existing attributes."""
        if name in dir(self) and name != 'reset':
            super().__setattr__(name, value)
        else:
            raise AttributeError(
                '"default" object has no attribute ' + repr(name))

    def reset(self):
        """Reset all attributes to their "factory default"."""
        vars(self).clear()


import sys as _sys
if not getattr(_sys.modules.get('sphinx'), 'SFS_DOCS_ARE_BEING_BUILT', False):
    # This object shadows the 'default' class, except when the docs are built:
    default = default()

from . import tapering
from . import array
from . import util
try:
    from . import plot2d
except ImportError:
    pass
try:
    from . import plot3d
except ImportError:
    pass

from . import fd
from . import td
